﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;
using System.Web;
using System.Web.Mvc;
using System.Web.Routing;

namespace MvcSettings.Caching
{
    [AttributeUsage(AttributeTargets.Class | AttributeTargets.Method, Inherited = true, AllowMultiple = false)]
    public class CacheOnServerAttribute : ActionFilterAttribute, IExceptionFilter
    {
        private readonly IOutputCacheManager _outputCacheManager;
        public int Duration { get; set; }
        public bool Sliding { get; set; }

        private string _cacheKey;

        public CacheOnServerAttribute()
        {
            _outputCacheManager = new OutputCacheManager();
            Duration = 60;
        }

        public CacheOnServerAttribute(IOutputCacheManager outputCacheManager)
        {
            if (outputCacheManager == null)
            {
                throw new ArgumentNullException("outputCacheManager");
            }

            _outputCacheManager = outputCacheManager;
            Duration = 60;
        }

        public override void OnActionExecuting(ActionExecutingContext filterContext)
        {
            _cacheKey = _outputCacheManager.GenerateCacheKey(filterContext);
            
            var cacheItem = _outputCacheManager.Get(_cacheKey) as OutputCacheItem;
            if (cacheItem != null)
            {
                filterContext.Result = new ContentResult()
                {
                    Content = cacheItem.Content,
                    ContentType = cacheItem.ContentType,
                };
                return;
            }
            else
            {
                // cache
                var cachingWriter = new StringWriter(CultureInfo.InvariantCulture);
                var originalWriter = filterContext.HttpContext.Response.Output;

                filterContext.HttpContext.Response.Output = cachingWriter;

                filterContext.HttpContext.Items[_cacheKey] = new Action<bool>(wasException =>
                {
                    // Removing this executing action from the context
                    filterContext.HttpContext.Items.Remove(_cacheKey);

                    // We restore the original writer for response
                    filterContext.HttpContext.Response.Output = originalWriter;

                    if (wasException)
                    {
                        return;
                    }

                    // Now we use owned caching writer to actually store data
                    var newCacheItem = new OutputCacheItem(
                        cacheKey: _cacheKey,
                        contentType: filterContext.HttpContext.Response.ContentType,
                        content: cachingWriter.ToString(),
                        controllerName: filterContext.RouteData.Values["controller"] as string,
                        actionName: filterContext.RouteData.Values["action"] as string,
                        areaName: filterContext.RouteData.DataTokens["area"] as string,
                        actionParameters: filterContext.ActionParameters,
                        cacheForSeconds: Duration,
                        sliding: Sliding
                    );

                    filterContext.HttpContext.Response.Write(newCacheItem.Content);

                    if (filterContext.HttpContext.Response.StatusCode == 200)
                    {
                        _outputCacheManager.Store(_cacheKey, newCacheItem, Sliding, Duration);
                    }
                });
            }
        }

        public void OnException(ExceptionContext filterContext)
        {
            CompleteChildAction(filterContext, true);
        }

        public override void OnResultExecuted(ResultExecutedContext filterContext)
        {
            bool wasException = filterContext.Exception != null;
            CompleteChildAction(filterContext, wasException);
        }

        private void CompleteChildAction(ControllerContext context, bool wasException)
        {
            var callback = context.HttpContext.Items[_cacheKey] as Action<bool>;

            if (callback != null)
            {
                callback.Invoke(wasException);
            }
        }
    }
}