﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using System.Web;
using System.Web.Mvc;
using System.Web.Routing;

namespace MvcSettings.Caching
{
    public class OutputCacheManager : IOutputCacheManager
    {
        private const string _cacheKeyPrefix = "_0utputCach3_";

        public IEnumerable<OutputCacheItem> CacheItems 
        { 
            get 
            {
                var cacheItems = new List<OutputCacheItem>();
                var items = _cache.Values.Where(key => key.Key.StartsWith(_cacheKeyPrefix));
                foreach (var item in items)
                {
                    var cacheItem = item.Value as OutputCacheItem;
                    if (cacheItem != null)
                    {
                        cacheItems.Add(cacheItem);
                    }
                }
                return cacheItems;
            } 
        }


        private readonly ICache _cache;

        public OutputCacheManager()
        {
            _cache = new AspNetCache();
        }

        public OutputCacheManager(ICache cache)
        {
            if (cache == null)
            {
                throw new ArgumentNullException("The cache argument is required");
            }

            _cache = cache;
        }

        public string GenerateCacheKey(ActionExecutingContext context)
        {
            string areaName = context.RouteData.DataTokens["area"] as string;
            string controllerName = context.RouteData.Values["controller"] as string;
            string actionName = context.RouteData.Values["action"] as string;
            var actionParameters = new RouteValueDictionary(context.ActionParameters);

            if (string.IsNullOrEmpty(areaName))
            {
                areaName = "";
            }

            return GenerateCacheKey(actionName, controllerName, areaName, actionParameters);
        }

        private string GenerateCacheKey(string actionName, string controllerName, string areaName, RouteValueDictionary routeValues)
        {
            if (string.IsNullOrEmpty(actionName))
            {
                throw new ArgumentNullException("actionName");
            }

            if (string.IsNullOrEmpty(controllerName))
            {
                throw new ArgumentNullException("controllerName");
            }

            if (areaName == null)
            {
                areaName = "";
            }

            // KEY FORMAT
            // CacheKeyPrefix*Area*Controller*Action*QueryString
            // Examples: (Main Area)(Home Controller)(About Action)(2 Parameters) 0utputcach3**home*about*param1=hello&param2=hello2
            // Examples: (Main Area)(Home Controller)(About Action)(No Parameters) 0utputcach3**home*about*

            var stringBuilder = new StringBuilder();
            stringBuilder.AppendFormat("*{0}", areaName.ToLower());
            stringBuilder.AppendFormat("*{0}", controllerName.ToLower());

            if (!string.IsNullOrEmpty(actionName))
            {
                stringBuilder.AppendFormat("*{0}*", actionName.ToLower());

                foreach (var parameter in routeValues)
                {
                    string value = parameter.Value == null ? "" : parameter.Value.ToString().ToLower();
                    stringBuilder.AppendFormat("{0}={1}#", parameter.Key.ToLower(), value);
                }

                // remove the last #
                stringBuilder.Remove(stringBuilder.Length - 1, 1);
            }

            using (SHA256 sHA = SHA256.Create())
            {
                string cacheKey = Convert.ToBase64String(sHA.ComputeHash(Encoding.UTF8.GetBytes(stringBuilder.ToString())));
                string cacheKeyWithPrefix = String.Concat(_cacheKeyPrefix, cacheKey.Replace("+", "-").Replace("/", "_").Replace("=", ""));
                return cacheKeyWithPrefix;
            }
        }

        public void Store(string cacheKey, object value, bool sliding, int secondsToCache)
        {
            _cache.Store(cacheKey, value, sliding, secondsToCache);
        }

        public object Get(string cacheKey)
        {
            var item = _cache.Get(cacheKey) as OutputCacheItem;
            if (item != null)
            {
                item.UpdateAccess();
            }
            return item;
        }

        /// <summary>
        /// Removes the cache entry with the cacheKey
        /// </summary>
        /// <param name="cacheKey"></param>
        public void Remove(string cacheKey)
        {
            _cache.Remove(cacheKey);
        }

        /// <summary>
        /// Clears all cached entries
        /// </summary>
        public void Clear()
        {
            var cacheKeys = _cache.Keys.Where(key => key.StartsWith(_cacheKeyPrefix));
            foreach (var cacheKey in cacheKeys)
                _cache.Remove(cacheKey);
        }

        /// <summary>
        /// Clears all cached entries for the given controller name
        /// </summary>
        /// <param name="controllerName">The controller to clear the cache</param>
        public void Clear(string controllerName)
        {
            var values = _cache.Values.Where(w => w.Key.StartsWith(_cacheKeyPrefix));
            foreach (var value in values)
            {
                var cacheItem = value.Value as OutputCacheItem;
                if (cacheItem != null && cacheItem.BelongsToController(controllerName))
                    _cache.Remove(value.Key);
            }
        }

        /// <summary>
        /// Clears all cached entries for the given controllers action
        /// </summary>
        /// <param name="actionName"></param>
        /// <param name="controllerName"></param>
        public void Clear(string actionName, string controllerName)
        {
            var values = _cache.Values.Where(w => w.Key.StartsWith(_cacheKeyPrefix));
            foreach (var value in values)
            {
                var cacheItem = value.Value as OutputCacheItem;
                if (cacheItem != null && cacheItem.BelongsToAction(controllerName, actionName))
                    _cache.Remove(value.Key);
            }
        }

        /// <summary>
        /// Clears all cached entries with the given action with parameters
        /// </summary>
        /// <param name="actionName"></param>
        /// <param name="controllerName"></param>
        /// <param name="routeValues"></param>
        public void Clear(string actionName, string controllerName, object routeValues)
        {
            var routeValueDictionary = new RouteValueDictionary(routeValues);
            string areaName = routeValueDictionary["area"] as string;

            if (routeValueDictionary.ContainsKey("area"))
                routeValueDictionary.Remove("area");

            string cacheKey = GenerateCacheKey(actionName, controllerName, areaName, routeValueDictionary);
            Remove(cacheKey);
        }
    }
}